#!/usr/bin/perl
#
# Copyright 2007, Fonality Inc. ALL RIGHTS RESERVED.
# Copyright (C) 2007  Fonality, Inc.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
#
# Author: Adrian Otto <support@fonality.com>
# Description: Utility for setting the IP address locally


$|++;
use strict;
use Term::ReadLine;
use Net::Ping;

use Getopt::Long;

# where most of the ip subs are:
require "/var/adm/bin/ip_funcs.pl";


my($interactive) = 0;
GetOptions ("i" => \$interactive);


my($hostname) = `/bin/hostname`; chomp($hostname);
my($tty) = `/usr/bin/tty`; chomp($tty);
my($RUNNING_LOCALLY) = 0;
if($tty =~ /^\/dev\/tty[1-6]$/) {
	$RUNNING_LOCALLY = 1;
}
else {
	if(! $interactive) {
		print STDERR "ERROR: Unable to run remotely without -i argument.\n";
		die "USAGE: $0 -i\n";
	}
}

# Make sure thie script only runs on pbxtraNNNN servers
unless($hostname =~ /^(?:pbxtra|trixbox|pbx)\d+$/) {
	die "This program can only be run on Fonality servers. Hostname=$hostname\n";
}

my ($mode, $settings) = show_current_config();

my $q   = "Do you want to change these settings? [y=yes, n=no]";
my $ans = _get_info_from_terminal($q);

if ($ans =~ /^y/i)
{
	change_config($mode, $settings);
}

if($RUNNING_LOCALLY) {
	print "Exiting...\n"; 
	sleep(3);
}

exit(0);

##############################################################################
# show_current_config
##############################################################################
sub show_current_config
{
	my $mode = shift;
	my %setting = %{ $_[0] || {}};

	$setting{ip} = _get_ip_address()    unless exists $setting{ip};
	$setting{sn} = _get_netmask()       unless exists $setting{sn};
	$setting{gw} = _get_default_route() unless exists $setting{gw};
	$setting{nw} = _get_network_addr($setting{ip})   unless exists $setting{nw};
	$setting{bc} = _get_broadcast_addr($setting{ip}) unless exists $setting{bc};

	my ($ns) = _get_nameservers();
	$setting{d1} = $ns->[0] unless exists $setting{d1};
	$setting{d2} = $ns->[1] unless exists $setting{d2};

	$mode = _get_bootproto() unless $mode;

	if($mode eq 'STATIC')
	{
		print "\nIP Address Mode: $mode\n\n";
		print "     IP Address: $setting{ip}\n";
		print "    Subnet Mask: $setting{sn}\n";
		print "Default Gateway: $setting{gw}\n";
		print "    Primary DNS: $setting{d1}\n";
		print "  Secondary DNS: $setting{d2}\n\n";
	}
	else
	{
		print "\nIP Address Mode: $mode (Auto-Assigned dynamically by DHCP server)\n\n";
	}

	return($mode, \%setting);
}

##############################################################################
# change_config
##############################################################################
sub change_config
{
	my ($mode, $setting) = @_;
	my $new_mode = 'DHCP';

	if($mode eq 'DHCP')
	{
		$q = "Assign a static IP address? [y=yes, n=no]";
		$ans = _get_info_from_terminal($q);
		if ($ans =~ /^y/i)
		{
			$setting = get_static_info($setting);
			$new_mode = 'STATIC';
			system("killall -9 dhclient");
		}
	}
	else
	{
		$q = "Activate DHCP? [y=yes, n=no]";
		$ans = _get_info_from_terminal($q);
		if($ans !~ /^y/i)
		{
			$setting = get_static_info($setting);
			$new_mode = 'STATIC';
		}
	}

	if (! confirm_config($new_mode, $setting))
	{
		print "Not making any changes.\n";
		return;
	}

	# Do the config stuff in one place so vtuns are started/stopped
	# correctly
	net_stop();
	if ($new_mode eq 'STATIC')
	{

		enable_static_ip($setting || {});
		net_start();
	}
	else
	{
		enable_dhcp_client($setting);
	}

	print "\nChanges complete.\n\n";
}

##############################################################################
# confirm_config
##############################################################################
sub confirm_config
{
	my ($mode, $setting) = @_;

	show_current_config($mode, $setting);
	my $q   = "Please review the configuration above. Apply settings now? [y=yes, n=no]";
	my $ans = _get_info_from_terminal($q);

	return $ans =~ /^y/i;
}

##############################################################################
# get_static_info
##############################################################################
sub get_static_info {
	my %setting = %{ $_[0] };

	my @q = (
		"     IP Address: (ip)",
		"    Subnet Mask: (sn)",
		"Default Gateway: (gw)",
		"    Primary DNS: (d1)",
		"  Secondary DNS: (d2)",
	);

	my $key;
	foreach my $q (@q)
	{
	    $q =~ /\(([^)]+)\)$/;
		$key = $1;
		
		if (exists $setting{$key} and $setting{$key})
		{
			# if they enter the ip of 192.168.1.100, 
			# lets recommend a gw of 192.168.1.1
			if ($key eq 'gw')
			{
				my $ip = $setting{ip};
				$ip =~ /^\s*(\d+\.\d+\.\d+)/;
				my $recommend_gateway = $1 . ".1";
				$setting{$key} = $recommend_gateway;
			}
		
			$q =~ s/$key/$setting{$key}/e;
		}
		else
		{
		    $q =~ s/$key/undefined/;
		}

	
          	my $ans = _get_info_from_terminal($q);

		# blank $ans signifies they're keeping the original value.
		if ($ans && !_is_ip($ans))
		{
			print "\nERROR: Please try again. That is not a valid IP Address.\n\n";
			redo;
		}

		# special cases
		if ($key eq 'ip')
		{
			if($setting{ip} ne $ans and !_is_available($ans))
			{
				print "\nERROR: Please try again. There is already a host using ip: $ans\n\n";
				redo;
			}
		}

		if ($key eq 'd1')
		{
			if ($setting{ip} eq $ans)
			{
				print "\nERROR: Please try again. Your Primary DNS cannot be the same as the ip.\n\n";
				redo;
			}
		}

		$setting{$key} = $ans if $ans;
	
    }

	# don't forget to calculate broadcast and network addresses (automagically)
	$setting{bc} = _calc_broadcast($setting{ip},$setting{sn});
	$setting{nw} = _calc_network($setting{ip},$setting{sn});

	return \%setting;
}

##############################################################################
# _get_info_from_terminal
##############################################################################
{
	my $term;

	sub _get_info_from_terminal {
		my $prompt = shift(@_) . ': ';

		if (not defined $term)
		{
			$term = new Term::ReadLine 'Fonality IP Address Utility';
		}

		my $answer = $term->readline($prompt);
		$term->addhistory($answer) if /\S/;

		return $answer;
	}
}

##############################################################################
# _check_settings: check network settings for sanity
# ip, gateway, broadcast address, network address, dns adress:
#		must be dotted quads where the first does not start with a zero
# subnet mask, 2nd dns address:
#		must be dotted quads
##############################################################################
sub _check_settings
{
	my $setting = shift;
	my $ns = _get_nameservers();

	# If these settings are blank, that also means that they're not an ip.
	if (!_is_ip($setting->{ip}, 'no_starting_zero'))
	{
		$setting->{ip} = _get_ip_address();
	}

	if (!_is_ip($setting->{gw}, 'no_starting_zero'))
	{
		$setting->{gw} = _get_default_route($setting->{ip});
	}

	if (!_is_ip($setting->{bc}, 'no_starting_zero'))
	{
		$setting->{bc} = _get_broadcast_addr($setting->{ip});
	}

	if (!_is_ip($setting->{nw}), 'no_start_ing_zero')
	{
		$setting->{nw} = _get_network_addr($setting->{ip});
	}

	if (!_is_ip($setting->{d1}, 'no_starting_zero'))
	{
		$setting->{d1} = $ns->[0];
	}

	# These can start with a zero
	if (!_is_ip($setting->{sn}))
	{
		$setting->{sn} = _get_netmask();
	}

	if (!_is_ip($setting->{d2}))
	{
		$setting->{d2} = $ns->[1];
	}

	return %$setting;
}
