#!/usr/bin/perl
use strict;

chdir("/var/adm/bin/");
use lib qw( /usr/local/rrdtool-1.2.15/lib/perl );
use lib qw( /usr/local/rrdtool-1.2.13/lib/perl );

# this variable is used for networking -- $KB bits in a kilobit
my $KB = 1000; # or 1024, 1000 is more typical in network measurement
my $DEBUG = 0;
my $pid_file = '/var/run/asterisk.pid';

system("/usr/bin/renice -1 $$ >>/dev/null 2>>/dev/null");

use RRDs;
use IO::Socket;
use Time::HiRes qw(usleep);
use Mail::Sender;
#use Data::Dumper;
#use Benchmark;

require("server_stats_alerts.pl");


# global sock
my $sock;

# other globals
our (%consecutive_ticks, %alreadyticked, %alarm_hash); 
my  ($stats_obj, $time, $last_raid_check, $last_call_quality_check);

my $hud_pid = get_hud_pid();


# create the tables
my $start = time - 2;
my $dst   = "AVERAGE";
my $DB    = "/var/lib/asterisk/server_stats.rrd";
my $DB2   = "/var/lib/asterisk/network_stats.rrd";


if(! -e $DB){
	RRDs::create(
		$DB, "--step=1",
		"DS:load:GAUGE:2:0:100",
		"DS:mem:GAUGE:2:0:100",	
		"DS:hdd:GAUGE:2:0:100",	
		"DS:ast:GAUGE:2:0:100",	
		"DS:hud:GAUGE:2:0:100",
		"RRA:$dst:0.5:1:3600",
		"RRA:$dst:0.5:360:60",
		"RRA:$dst:0.5:1440:60",
		"RRA:$dst:0.5:10080:60",
		"RRA:$dst:0.5:43200:60",
	) or die "Cannot create rrd (" . RRDs::error . ")";
}


if(! -e $DB2){
	RRDs::create(
		$DB2, "--step=1",
		"DS:rx:GAUGE:2:0:104857600",
		"DS:tx:GAUGE:2:0:104857600",
		"RRA:$dst:0.5:1:3600",
		"RRA:$dst:0.5:360:60",
		"RRA:$dst:0.5:1440:60",
		"RRA:$dst:0.5:10080:60",
		"RRA:$dst:0.5:43200:60",
	) or die "Cannot create rrd (" . RRDs::error . ")";
}


# debug mode
if (!$DEBUG)
{
	my $pid = fork();
	die "running in bg ($pid)\n" if $pid;

	close(STDIN);
	close(STDERR);
	close(STDOUT);
}

# used as static variables for The Big Loop
my ($num, $rx, $tx, $old_tx, $old_rx, $tx_diff, $rx_diff, $conf, $kbhalf);


while(usleep(1000000))
{
	my $newtime = time();

	# if last time < now
	if ($time != $newtime)
	{
		$time = $newtime;

		# get our config if we have new info
		$conf = get_conf($conf);

		# Add network info to the chart.
		# Complication: when we poll too quickly, there's no new
		# data. We need to poll about every 2 seconds to make sure
		# there is new data. Otherwise, we end up with one second with
		# 0 traffic, and the next second with 2 seconds of traffic.
		#
		# To get around this, if we see 0 change in traffic, we'll simply
		# apply the next second's traffic in half to this second and the
		# next second.
		($rx, $tx) = get_net_info($conf);
		if ($old_tx && $old_rx)
		{
			# diff in time over 1 second
			$tx_diff = ($tx - $old_tx);
			$rx_diff = ($rx - $old_rx);

			# did we not change?
			if ($tx_diff == 0 && $rx_diff == 0)
			{
				$kbhalf = 1;
			}

			# amount of traffic changed
			else
			{
				# set old amounts to compare with now
				$old_tx = $tx;
		       	$old_rx = $rx;

				# should we apply this in half over 2 seconds?
				if ($kbhalf)
				{
					$rx_diff = round($rx_diff / 2); # divide in half and round
					$tx_diff = round($tx_diff / 2); # divide in half and round

					# update previous second
					unless (RRDs::update($DB2, ($time-1).":$rx_diff:$tx_diff"))
					{
						print STDERR "Cannot update rrd (" . RRDs::error . "$!)\n";
					}

					$kbhalf = 0; # don't worry about halfing next time
				}

				# do a regular update
				unless (RRDs::update($DB2, "$time:$rx_diff:$tx_diff"))
				{
					print STDERR "Cannot update rrd (" . RRDs::error . "$!)\n";
				}
			}
		}
		else
		{
			# set old amounts to compare with now
			$old_tx = $tx;
		    $old_rx = $rx;
		}

		# Initialize the data from proc files via module.
		my $stats = init_stats($stats_obj);

		my $cpu = get_stat($stats, 'cpu');
		my $ast_mem = get_asterisk_mem() / get_stat($stats, 'memtotal') * 100;		

		my ($hdd_total, $hdd_free);
		foreach my $hdd (values %{get_stat($stats, 'diskusage')}){
			$hdd_total += $hdd->{'total'};
			$hdd_free += $hdd->{'free'};
		}

		my $percent_hdd = (100-(($hdd_free/$hdd_total) * 100));		

		my $hud_cpu = get_hud_cpu();
		$hud_cpu = $cpu if($hud_cpu > $cpu);		

		my $mem = (get_stat($stats, 'swapused')/(512*1024)) * 100;
		
		$cpu         = round($cpu);
		$mem         = round($mem);
		$percent_hdd = round($percent_hdd);
		$ast_mem     = round($ast_mem);
		$hud_cpu     = round($hud_cpu);

		#print "$time:$cpu:$mem:$percent_hdd:$ast_mem:$hud_cpu\n";
		# update the "resources" RRD table
		unless (RRDs::update($DB, "$time:$cpu:$mem:$percent_hdd:$ast_mem:$hud_cpu"))
		{
			print STDERR "Cannot update rrd (" . RRDs::error . "$!)\n";
		}
	
		if(($cpu >= $conf->{cpu}->{value} && defined($conf->{cpu}->{value})) || (($cpu > 50) && $DEBUG))	
		{
			my @ltime = localtime($time);

			# don't send CPU alarms at 4am, this is when daily crons run
			if ($ltime[2] != 4)
			{
				send_alarm($conf->{cpu}->{email},"CPU",$cpu,'cpu', $conf);
			}
		}
				
		if(($hud_cpu >= $conf->{hudcpu}->{value} && defined($conf->{hudcpu}->{value})) || (($hud_cpu > 50) && $DEBUG))	
		{
			send_alarm(
				$conf->{hudcpu}->{email},
				"HUD CPU",
				"$hud_cpu%",
				'hudcpu',
				$conf,
			);
		}
		
		if(($mem >= $conf->{mem}->{value} && defined($conf->{mem}->{value})) || (($mem > 15) && $DEBUG))	
		{
			send_alarm(
				$conf->{mem}->{email},
				"swap memory",
				"$mem%",
				'mem',
				$conf,
			);
		}
		
		if(($percent_hdd >= $conf->{hdd}->{value} && defined($conf->{hdd}->{value})) || (($percent_hdd > 50) && $DEBUG))
		{
			send_alarm(
				$conf->{hdd}->{email},
				"hard drive", 
				"$percent_hdd%", 
				'hdd',
				$conf,
			);
		}

		if(($ast_mem >= $conf->{astmem}->{value} && defined($conf->{astmem}->{value})) || (($ast_mem > 50) && $DEBUG))	
		{
			send_alarm(
				$conf->{astmem}->{email},
				"Asterisk memory",
				"$ast_mem%", 
				'astmem',
				$conf,
			);
		}

		if((($rx_diff*8/$KB) >= $conf->{rx}->{value} && defined($conf->{rx}->{value})) || (($rx_diff*8 > (30*$KB)) && $DEBUG))	
		{
			send_alarm(
				$conf->{rx}->{email},
				"kb/sec received",
				($rx_diff * 8 / $KB), 
				'rx',
				$conf,
			);
		}

		if((($tx_diff*8/$KB)>= $conf->{tx}->{value} && defined($conf->{tx}->{value})) || (($tx_diff*8 > (30*$KB)) && $DEBUG))	
		{
			send_alarm(
				$conf->{tx}->{email},
				"kb/sec sent",
				($tx_diff * 8 / $KB), 
				'tx',
				$conf,
			);
		}

		if ($last_raid_check <= $time - 60)
		{
			$last_raid_check = $time;

			my $raid_status = get_raid_status();
			if (defined $conf->{'raid'}->{'value'} &&
				(
					($conf->{'raid'}->{'value'} =~ /$raid_status/i)
					||
					($conf->{'raid'}->{'value'} =~ /Any/i && $raid_status !~ /normal/i)
				))
			{
				send_alarm(
					$conf->{'raid'}->{'email'},
					'Raid Failure',
					$raid_status,
					'raid',
					$conf,
					60,
				);
			}
		}

		if ($last_call_quality_check <= $time - 60)
		{
			$last_call_quality_check = $time;

			my $call_quality_status = get_call_quality_status();
			
			if (defined $conf->{'call_quality'}->{'enable'})
			{
				send_alarm(
					$conf->{'call_quality'}->{'email'},
					'Call Quality Issue',
					$call_quality_status,
					'call_quality',
					$conf,
					60,
				);
			}
		}


		# this is used for Sangoma cards, as they have a different method for polling
		# done before the loop so it doesn't depend on the order the loop processes cards
		# have to reload every time through to make sure we catch conf changes
		my ($vcard, $chanspan) = (1, 0);
		foreach my $t1 (
			grep { exists $conf->{$_}{'type'} && $conf->{$_}{'type'} =~ /t1/i }
			grep /card/,
			keys %$conf)
		{
			$chanspan += $conf->{$t1}{'chanspan'} || 0;
			$vcard++;
		}

		# It's possible to have unplugged/unassigned ports/cards and we don't want to log all those errors.
		# So only loop over the cards in the conf file.
		foreach my $pri_name (sort grep /card/, keys %$conf)
		{
			next unless exists $conf->{$pri_name}{'enable'} && $conf->{$pri_name}{'enable'};

			# make a shortcut for this oft-used value
			my $cond = $conf->{$pri_name}{'value'};

			my $status = get_card_status(
				$conf->{$pri_name}{'type'},
				$conf->{$pri_name}{'chan'},
				$vcard,
				$chanspan,
			);

			if ($conf->{$pri_name}->{'type'} eq 't1')
			{
				if ($status =~ /$cond/i or ( $cond =~ /any/i && $status =~ /red|yellow|down/i) )
				{
					send_alarm(
						$conf->{$pri_name}->{'email'},
						"$pri_name status alarm",
						$status,
						$pri_name,
						$conf,
					);
				}
			}

			# analog card
			else
			{
				next if not defined $status;    # unable to probe card properly, received undef
				next if $status =~ /normal/i;   # if it's normal, no need to alert
				if ($status =~ /$cond/i or ( $cond =~ /any/i && $status !~ /normal/i ) )
				{
					send_alarm(
						$conf->{$pri_name}{'email'},
						"$pri_name voltage alarm",
						$status,
						$pri_name,
						$conf
					);
				}
			}
		}
	}
}
close($sock);

sub get_asterisk_pid
{
	my $pid = -1;

	if (-f $pid_file)
	{
		if (open my $ast_pid, '<', $pid_file)
		{
			chomp($pid = <$ast_pid>);
		}
	}

	return $pid;
}

# $ast_pid and $last_pidm are used as static variables in get_asterisk_mem
my ($ast_pid, $last_pidm);
sub get_asterisk_mem
{
	my $process_vsize;
	my $pidm = (stat($pid_file))[9] ;

	if($pidm != $last_pidm)
	{
		$ast_pid = get_asterisk_pid();
		$last_pidm = $pidm;
	}

	# read file and get 22nd column
	if(-f "/proc/$ast_pid/stat") {
		open(STAT, "/proc/$ast_pid/stat");
		my @stats = split / /, <STAT>;
		close(STAT);

		$process_vsize = int($stats[22] / 1024); # Convert to MB
	}

	return $process_vsize;
}

# $cputime and $proctime are used as static variables in get_hud_cpu
my ($cputime, $proctime);
sub get_hud_cpu
{
	my $percent = 0;
	my $current_cputime  = get_cpu_time();
	my $current_proctime = get_hud_cpu_time();
	
	if ($current_cputime <= $cputime)
	{
		$percent = 0;
	}
	else
	{
		
		# (this process's cputime - last process's cputime) over (entire cputime - last cputime)
		$percent = 100 * (($current_proctime - $proctime) / ($current_cputime - $cputime));
		$percent = "99.9" if $percent > 99.9;
	}
	$proctime = $current_proctime;
	$cputime  = $current_cputime;
	
	return $percent;
}

# look at proc files for hud server
sub get_hud_pid
{
	opendir(PROC, "/proc") || return undef;
	foreach my $pid (grep { /^\d+$/ } readdir(PROC))
	{
		open(CMD, "</proc/$pid/stat") || next;
		my $stats = <CMD>;
		close(CMD);

		if(index($stats, 'saxirc') >= 0)
		{
			return $pid;
		}
	}
}


# number of jiffies (1/100ths of a second) that the system spent
# in user mode, user mode with low priority (nice), system mode,
# and the idle task summed up
sub get_cpu_time
{
	my %stat;

	open PROC, "</proc/stat";
	@stat{"cpu", "user", "nice", "sys", "idle"} = split(/\s+/, <PROC>);
	close(PROC);

	return $stat{user} + $stat{nice} + $stat{sys} + $stat{idle};
}


# jiffies that this process has been scheduled in user mode +
# jiffies that this process has been scheduled in kernel mode
sub get_hud_cpu_time
{
	my %stat;

	if((! -e "/proc/$hud_pid/stat") || !$hud_pid)
	{
		$hud_pid = get_hud_pid();	
	}

	open PROC, "</proc/$hud_pid/stat";
	@stat{"pid", "comm", "state", "ppid", "pgrp", "session", "tty_nr", "tpgid", "flags",
	"minflt", "cminflt", "majflt", "cmajflt", "utime", "stime", "cutime",
	"cstime", "priority", "nice", "who_cares", "itrealvade",
	"starttime", "vsize", "rss", "rlim", "startcode", "endcode", "startstack",
	"kstkesp", "kstkeip", "signal", "blocked", "sigignore", "sigcatch",
	"wchan", "nswap", "cnswap", "exit_signal", "processor"} = split(" ", <PROC>);
	close(PROC);

	return $stat{stime} + $stat{utime};		# + $stat{cutime};
}

sub get_net_info
{
	my $conf = shift;

	# specified device, otherwise read eth0
	my $dev = $conf->{netdev} || "eth0";

	my @net_info;
	my ($tx, $rx);
	open(FILE, "/proc/net/dev");
	while(<FILE>)
	{
		if($_ =~ m/^\s*$dev:(.*?)$/){
			@net_info = split(/\s+/,$1);
			($rx, $tx) = @net_info[0,8];
			last;
		}
	}
	return ($rx, $tx);
}

sub get_card_status
{
	my ($type, $chan) = (shift, shift);
	my $vcard         = shift || 1;
	my $chanspan      = shift || 0;

	if ($type =~ /t1/i)
	{
		# First check for red or yellow alarms
		my $alarms = get_t1_alarms($chan);
		return 'red'    if grep /red/i, @$alarms;
		return 'yellow' if grep /yellow/i, @$alarms;

		# Check to make sure the card is up at all
		my $status = `asterisk -r -x "pri show span $chan"`;
		return 'down' if $status =~ /status:.*?down/i;

		# all good! (as far as we know)
		return 'none';
	}

	my $voltage;
	my $response = `asterisk -r -x "zap show regdump $chan"` ;

	if ($response =~ /TIP:\s*-?(\d+)/)
	{
		$voltage = $1;
	}
	else
	{
		# let's try sangoma way
		$response = `/usr/sbin/wanpipemon -i w${vcard}g1 -c astats -m @{[ $chan - $chanspan ]}`;
		if ($response =~ /VOLTAGE\s*:\s*(\d+)/)
		{
			$voltage = $1;
		}
		else
		{
			return undef;
		}
	}

	return	$voltage < 3  ? 'no'     :
			$voltage < 13 ? 'normal' :		# on call
			$voltage < 35 ? 'low'    :
							'normal' ;		# idle
}

sub get_t1_alarms
{
	my $span = shift;
	my %alarms = (
		none            => 0,
		recover         => 1,
		loopback        => 2,
		yellow          => 4,
		red             => 8,
		blue            => 16,
		notopen         => 32,
	);

	my $ZT_SPANSTAT = -1066382838;
	my $spaninfo = zt_spaninfo({ spanno => $span });

	open(my $ctl, '+<', '/dev/zap/ctl') || return "Couldn't open /dev/zap/ctl for rw: $!";
	ioctl($ctl, $ZT_SPANSTAT, $spaninfo);
	close($ctl);

	my $tmp = { zt_spaninfo_unp($spaninfo) };
	return [ grep { $tmp->{alarms} & $alarms{$_} } keys %alarms ];

	sub zt_spaninfo
	{
		my %tmp = %{$_[0]};
		return
		pack("iA20A40iiiiiiiiiiii",
			$tmp{spanno},
			$tmp{name},
			$tmp{desc},
			$tmp{alarms},
			$tmp{txlevel},
			$tmp{rxlevel},
			$tmp{bpvcount},
			$tmp{crc4count},
			$tmp{ebitcount},
			$tmp{fascount},
			$tmp{irqmisses},
			$tmp{syncsrc},
			$tmp{numchans},
			$tmp{totalchans},
			$tmp{totalspans},
		);
	}

	sub zt_spaninfo_unp
	{
		my $tmp = shift;
		my @tmp = unpack("iA20A40iiiiiiiiiiii", $tmp);
		my $i = 0;

		return map { $_ => $tmp[$i++] }
			qw/spanno name desc alarms txlevel rxlevel bpvcount crc4count
			ebitcount fascount irqmisses syncsrc numchans totalchans totalspans/;
	}
}

sub get_raid_status
{
	my $info;

	if (-e '/usr/sbin/hpacucli' && -x _)
	{
		# this interprets "rebuilding" status as normal;
		# by the time a drive is rebuilding, the user already knows something has failed
		$info = `/usr/sbin/hpacucli ctrl all show config detail`;
		return 'fail' if( $info =~ /failed/i );
		return 'normal';
	}
	else
	{
		$info = `/bin/cat /proc/mdstat`;

		foreach (split /\n/, $info)
		{
			if (/block/)
			{
				my ($status) =	/^
									[^\/]+ \/		# look for a slash, then match it
									[^\[]+			# look for a '[' after that
									\[ ([^\]]+) \]	# match everything inside brackets (status indicators)
								/x;

				return 'fail'    if $status =~ /F/i;
				return 'missing' if $status =~ /_/;
			}
		}

		return 'normal';
	}
}

sub get_call_quality_status()
{
	my $zaptel_file = '/etc/zaptel.conf';
	my $timing;

	if (-e $zaptel_file)
	{
		open(ZAPTEL, "<$zaptel_file");
		my @lines = <ZAPTEL>;
		close(ZAPTEL);

		my $timing_sources_found = 0;
		my $span_lines_found = 0;
		my $analog_lines_found = 0;
		my $span;

		foreach my $line (@lines)
		{
			chomp($line);

			if ($line =~ /^\s*span=(\d),(\d),/)
			{
				$timing->{$1} = $2;

				$span_lines_found++;

				if ($2 == 1)
				{
					$span = $1;
					$timing_sources_found++;
				}
			}
			elsif ($line =~ /^\s*fx[so]/)
			{
				$analog_lines_found++;
			}
		}

		# We found a span, but found no timing, or more then one timing source, complain
		if ($span_lines_found > 0 && $timing_sources_found < 1)
		{
			return "no spans are currently marked as a timing source";
		}

		if ($span_lines_found > 0 && $timing_sources_found > 1)
		{
			return "more then one span is marked as a timing source";
		}
	
		# SIP only system with no cards is going to be a huge call quality issue
		unless ($analog_lines_found || $span_lines_found)
		{
			return "due to the lack of any timing source";
		}

        # In this scenario only one span was found, let's make sure it's not in a state of red or down
        if ($span_lines_found == 1 && $timing_sources_found == 1)
        {
			# First check for red or yellow alarms
			my $alarms = get_t1_alarms($span);

			if (grep /red/i, @$alarms)
			{
				return "span $span being in red alarm state";
			}
        }

	}
}

# This closure wraps around the methods that call the linux stats package to cache the results safely.
# The preferred alternative would be to write an object that would keep this information.
# By using this instead it keeps things lighter-weight. (it's a fly-weight pseudo-object)
{
	my ($_stats_pkg, %_stats_pkg_opts, $_init_method, $_lower_keys);

	BEGIN
	{
		# This defines the options used for each package, including:
		#	opts: the options to use when creating a new object of this package
		#	init: the method name of the method to use when initializing the stats for a package
		#	locs: the location in the hash of the various information we want access to
		#			note: the keys are kept in an array and traversed down in a loop (see sub 'get_stat')
		#	ordr: the order in which to attempt to load each package. Lower is better.
		#			note: currently using 4 and 8 to give breathing room for future expansions (instead of 0 and 1)
		#
		# This is wrapped in a BEGIN so that the assignment happens before the opening code. Otherwise it would
		# be assigned out of sequence.
		%_stats_pkg_opts =
		(
			'Sys::Statistics::Linux' =>
			{
				ordr => 4,
				opts => 'ProcStats => 1, MemStats => 1, DiskUsage => 1, CpuStats => 1',
				init => 'get',
				locs =>
				{
					cpu 		=> [ 'cpustats', 'cpu', 'total' ],
					memtotal	=> [ 'memstats', 'memtotal' ],
					diskusage	=> [ 'diskusage' ],
					swapused	=> [ 'memstats', 'swapused' ],
				},
			},

			'Linux::Statistics' =>
			{
				ordr => 8,
				opts => 'ProcStats => 1, MemStats => 1, DiskUsage => 1',
				init => 'getStats',
				locs => {
					cpu 		=> [ 'procstats', 'total' ],
					memtotal	=> [ 'memstats', 'memtotal' ],
					diskusage	=> [ 'diskusage' ],
					swapused	=> [ 'memstats', 'swapused' ],
				},
			},
		);

		# Recursively descend through however many keys there are and make sure that all keys are lower case.
		# Different modules use different cases but all keys are unique regardless of case (known from inspection)
		#
		# In order to make sure this method is compiled in time, it needs to also be in a begin block.
		$_lower_keys = sub {
			my $hash_ref = shift;

			foreach my $key (keys %$hash_ref) {
				if (ref($hash_ref->{$key}) eq 'HASH')
				{
					$_lower_keys->($hash_ref->{$key});
				}

				my $lkey = lc $key;
				if ($lkey ne $key)
				{
					$hash_ref->{$lkey} = delete $hash_ref->{$key};
				}
			}

			return $hash_ref;
		};
	}

	# The modules are loaded with different options and the module loaded needs to be cached.
	# Check which module we load (prefer Sys::Statistics::Linux over Linux::Statistics) and
	# load applicable options.
	sub load_stats_pkg
	{
		foreach my $pkg ( sort { $_stats_pkg_opts{$a}{'ordr'} <=> $_stats_pkg_opts{$b}{'ordr'} } keys %_stats_pkg_opts)
		{
			eval "use $pkg";
			if (! $@)
			{
				$stats_obj = eval "$pkg->new( $_stats_pkg_opts{$pkg}{'opts'} )";
				$_stats_pkg = $pkg;
				last;
			}
		}

		if (not defined $_stats_pkg)
		{
			die "Could not load any Linux statistics package.\n";
		}
	}

	# Calls the (package-dependant) appropriate stat init function and returns a hash ref that has all keys lc()'ed.
	# It's important that no eval's are used in the middle of the loop because they're prohibitively computationally
	# expensive, so instead we use the UNIVERSAL can method to locate the method and call it directly. This has the
	# added benefit that we can determine if something has gone wrong and we don't have the object we thought we did.
	#
	# The method is cached, however we still have to call $_lower_keys->() every time through the loop since
	# we're given a new data object with every call. That's the single largest performance trade-off for supporting
	# multiple modules.
	sub init_stats
	{
		my $obj  = shift;

		if (defined $_init_method)
		{
			return $_lower_keys->($obj->$_init_method);
		}

		$_init_method = $obj->can($_stats_pkg_opts{$_stats_pkg}{'init'});

		if (not defined $_init_method)
		{
			die "Could not initialize statistics for stats package $_stats_pkg\n";
		}

		return $_lower_keys->($obj->$_init_method);
	}

	# Differences between stat packages are remediated here.
	# This is the defining function that this entire closure was predicated on.
	#
	#	NOTE: if you get the "seeky" reference then i salute you, fellow geek.
	#
	# It merely creates a reference into the data object and returns the reference; it does not deep copy.
	# This means that despite the minor slowdown it's almost as fast as having direct access to the data.
	sub get_stat
	{
		my $stats = shift;
		my $seeky = shift;

		# Get a reference to the stats.
		my $ret = $stats;

		# Loop through the keys and set the reference pointer to the proper location.
		foreach my $key ( @{ $_stats_pkg_opts{$_stats_pkg}{'locs'}{$seeky} } )
		{
			$ret = $ret->{$key};
		}

		# Return the reference.
		return $ret;
	}
}

# make sure the use statements fire before everything else as appropriate
BEGIN { load_stats_pkg() }
